//
//  N80View.m
//  N80
//
//  Created by Pecos on 2012/08/25.
//  Copyright (c) 2012年 H.Kanemoto. All rights reserved.
//

#import "N80View.h"

@implementation N80View

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code here.
    }
    return self;
}
- (id)initWithFrame:(NSRect)frameRect pixelFormat:(NSOpenGLPixelFormat*)format
{
	// xibにViewを配置しているときはinitWithFrameが呼ばれないので注意
    self = [super initWithFrame:frameRect];
    if (self) {
        // initialize
	}
    return self;
}
- (void)awakeFromNib
{
	CGDataProviderRef dp = CGDataProviderCreateWithCFData((CFDataRef)[NSData dataWithContentsOfFile:
		[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/N80tile.png"]]);
	tile = CGImageCreateWithPNGDataProvider(dp, NULL, true, kCGRenderingIntentDefault);
}
- (void)reshape
{
    int width = [self frame].size.width;
    int height = [self frame].size.height;
	width_scale = (float) width / (float) height;
	NSLog(@"width scale = %f", width_scale);
    // view port transform
	if (width == 640) {
 	   glViewport(0, 0, width, height);
	} else {
		int w = width, h = w * 10 / 16;
		if (h > height) {
			h = height;
			w = h * 16 / 10;
		}
		int dx = 0, dy = 0;
		if (w < width)
			dx = (width - w) / 2;
		if (h < height)
			dy = height - h;
		glViewport(dx, dy, w, h);
	}
    // projection transform
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	// orthoscopic view matrix
	//glOrtho(width_scale*-1, width_scale, -1.0f, 1.0f, -1.0f, 1.0f);
	glOrtho(-1, 1, -1.0f, 1.0f, -1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// disabling the depth test
	glDisable(GL_DEPTH_TEST);
}
- (void)drawRect:(NSRect)dirtyRect
{
	// Drawing code here.
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f); // default background color
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    // View Transform
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    // Modeling Transform
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    [self drawScene];
    glFlush();
/*
    CGContextRef myContext = [[NSGraphicsContext currentContext] graphicsPort];
	if (self.n80 != NULL && !self.n80.pause) {
		CGImageRef img = [self.n80 reqImage];
		if (img != NULL) {
			CGRect r = CGRectMake(0, 0, 640, 400);
			CGContextDrawImage(myContext, r, img);
			CGImageRelease(img);
			return;
		}
	}
	CGRect r = CGRectMake(0, 0, 128, 128);
	for (int w = 0; w < 640; w += 128) {
		for (int h = 0; h < 400; h += 128) {
			r.origin.x = w;
			r.origin.y = h;
			CGContextDrawImage(myContext, r, tile);
		}
	}*/
}
- (void)drawScene
{
	if (self.n80 == NULL)
		return;
	if (backTexture == 0) {
		void *bitmap = [self.n80 reqTexBitmap];
		if (bitmap == NULL) {
			return;
		}
		[self loadTextureName:&backTexture :bitmap];
	}
	// update texture image
	glBindTexture(GL_TEXTURE_2D, backTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 640, 400, 0, GL_RGBA, GL_UNSIGNED_BYTE, [self.n80 reqTexBitmap]);
	// clear screen
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	// draw texture
	[self drawTexture];
}
- (void)drawTexture {
	float vertices[] = {
		-1.0f, -1.0f,		//left top
		-1.0f,  1.0f,		//left bottom
		1.0f, -1.0f,		//right top
		1.0f,  1.0f,		//right bottom
	};
	float texcoords[] = {
		0.0f, 0.0f,			//left top
		0.0f, 1.0f,			//left bottom
		1.0f, 0.0f,			//right top
		1.0f, 1.0f,			//right bottom
	};
	GLubyte colors[] = {
		255, 255, 255, 255,
		255, 255, 255, 255,
		255, 255, 255, 255,
		255, 255, 255, 255
	};
	glActiveTexture(GL_TEXTURE0);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(2, GL_FLOAT, 0, vertices);
	glColorPointer(4, GL_UNSIGNED_BYTE, 0, colors);
	glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
	
	// draw
	glBindTexture(GL_TEXTURE_2D, backTexture);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glBindTexture(GL_TEXTURE_2D, 0);
}
- (void)loadTextureName:(GLuint *)texture :(void *)bitmap {
	// create texture
	glGenTextures(1, texture);
	// bind texture
	glBindTexture(GL_TEXTURE_2D, *texture);
	// set image data to texture
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 640, 400, 0, GL_RGBA, GL_UNSIGNED_BYTE, bitmap);
	// configure texture
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glEnable(GL_TEXTURE_2D);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
}
- (BOOL)acceptsFirstResponder {
	return YES;
}
- (void)keyDown:(NSEvent *)theEvent {
	//NSLog(@"KeyDown %d",[[theEvent characters] characterAtIndex:0]);
	if (self.n80 != nil)
		[self.n80 OnKeyDown:theEvent];
}
- (void)keyUp:(NSEvent *)theEvent {
	//NSLog(@"KeyUp %d",[[theEvent characters] characterAtIndex:0]);
	if (self.n80 != nil)
		[self.n80 OnKeyUp:theEvent];
}
- (void)flagsChanged:(NSEvent *)theEvent {
	//NSLog(@"Flags Changed");
	if (self.n80 != nil)
		[self.n80 OnFlagsChanged:theEvent];
}
- (void)dealloc
{
	CGImageRelease(tile);
	[super dealloc];
}
@end
