//
//  n80em.m
//  N80
//
//  Created by Pecos on 2012/08/22.
//  Copyright (c) 2012年 H.Kanemoto. All rights reserved.
//

#import "n80em.h"
#include <mach/mach_time.h>
#include <AudioUnit/AudioUnit.h>
//-----------------------------------------------------------------------------
//	Emulator C Functions
//-----------------------------------------------------------------------------
#define NOTDONE	0
#define DONE	1
#define STDSIZE	(32576)
#define MAXSIZE	(32768)
#define SP_VAL	(0x7f3e)
#define PC_VAL	(0xff3d)
#define SAMPLE_RATE (44100.0)

static N80LX		n80lx;
static NSConditionLock *syncSound = NULL;
static SAMPLETsBANK	beep_bank;
static SAMPLETsBANK	sound_bank;
static const DWORD	SPITZ_COLOR[8] = {0xFF000000,0xFFff0000,0xFF00ff00,0xFFffff00,0xFF0000ff,0xFFff00ff,0xFF00ffff,0xFFffffff};
static int 			bltmode = 0;
static BYTE  		down_matorix[80][2] =
{
	{0,0xfe}, {0,0xfd}, {0,0xfb}, {0,0xf7},
	{0,0xef}, {0,0xdf}, {0,0xbf}, {0,0x7f},
	{1,0xfe}, {1,0xfd}, {1,0xfb}, {1,0xf7},
	{1,0xef}, {1,0xdf}, {1,0xbf}, {1,0x7f},
	{2,0xfe}, {2,0xfd}, {2,0xfb}, {2,0xf7},
	{2,0xef}, {2,0xdf}, {2,0xbf}, {2,0x7f},
	{3,0xfe}, {3,0xfd}, {3,0xfb}, {3,0xf7},
	{3,0xef}, {3,0xdf}, {3,0xbf}, {3,0x7f},
	{4,0xfe}, {4,0xfd}, {4,0xfb}, {4,0xf7},
	{4,0xef}, {4,0xdf}, {4,0xbf}, {4,0x7f},
	{5,0xfe}, {5,0xfd}, {5,0xfb}, {5,0xf7},
	{5,0xef}, {5,0xdf}, {5,0xbf}, {5,0x7f},
	{6,0xfe}, {6,0xfd}, {6,0xfb}, {6,0xf7},
	{6,0xef}, {6,0xdf}, {6,0xbf}, {6,0x7f},
	{7,0xfe}, {7,0xfd}, {7,0xfb}, {7,0xf7},
	{7,0xef}, {7,0xdf}, {7,0xbf}, {7,0x7f},
	{8,0xfe}, {8,0xfd}, {8,0xfb}, {8,0xf7},
	{8,0xef}, {8,0xdf}, {8,0xbf}, {8,0x7f},
	{9,0xfe}, {9,0xfd}, {9,0xfb}, {9,0xf7},
	{9,0xef}, {9,0xdf}, {9,0xbf}, {9,0x7f}
};
static BYTE  		up_matorix[80][2] =
{
	{0,0x01}, {0,0x02}, {0,0x04}, {0,0x08},
	{0,0x10}, {0,0x20}, {0,0x40}, {0,0x80},
	{1,0x01}, {1,0x02}, {1,0x04}, {1,0x08},
	{1,0x10}, {1,0x20}, {1,0x40}, {1,0x80},
	{2,0x01}, {2,0x02}, {2,0x04}, {2,0x08},
	{2,0x10}, {2,0x20}, {2,0x40}, {2,0x80},
	{3,0x01}, {3,0x02}, {3,0x04}, {3,0x08},
	{3,0x10}, {3,0x20}, {3,0x40}, {3,0x80},
	{4,0x01}, {4,0x02}, {4,0x04}, {4,0x08},
	{4,0x10}, {4,0x20}, {4,0x40}, {4,0x80},
	{5,0x01}, {5,0x02}, {5,0x04}, {5,0x08},
	{5,0x10}, {5,0x20}, {5,0x40}, {5,0x80},
	{6,0x01}, {6,0x02}, {6,0x04}, {6,0x08},
	{6,0x10}, {6,0x20}, {6,0x40}, {6,0x80},
	{7,0x01}, {7,0x02}, {7,0x04}, {7,0x08},
	{7,0x10}, {7,0x20}, {7,0x40}, {7,0x80},
	{8,0x01}, {8,0x02}, {8,0x04}, {8,0x08},
	{8,0x10}, {8,0x20}, {8,0x40}, {8,0x80},
	{9,0x01}, {9,0x02}, {9,0x04}, {9,0x08},
	{9,0x10}, {9,0x20}, {9,0x40}, {9,0x80}
};
static void chg_bltfunc(int mode) {
	// mode: 0 = 40x20, 1 = 40x25, 2 = 80x20, 3 = 80x25
	bltmode = mode;
}
static void enter() {
	// spin lock
	while (!(OSAtomicCompareAndSwap32(0, 1, &n80lx.vram_sync)))
		;
}
static void leave() {
	n80lx.vram_sync = 0;
}
static inline void append_tick_to_array(LPSAMPLETsBANK bank, DWORD new_state)
{
	// 現在のTick累積を配列に追加する
	LPSAMPLETsARR arr = &bank->banks[bank->bank];
	if (arr->count < MAXSTsARRAYS - 1) {
		// CPU STATE数をALSA frame数に変換して累積する
		double frames = n80lx.frame_keisu * bank->current.ticks;
		arr->arr[arr->count].ticks   = frames;
		arr->arr[arr->count++].state = bank->current.state;
		arr->ticks_sum += frames;
	} else
		fprintf(stderr, "beep ticks array overflow\n");
	
	bank->current.ticks = 0;
	bank->current.state = new_state;
}
static void _n80_beep_on() {
	[syncSound lock];
	if (beep_bank.current.state == 0) {
		append_tick_to_array(&beep_bank, 1);
	}
	[syncSound unlock];
}
static void _n80_beep_off() {
	[syncSound lock];
	if (beep_bank.current.state == 1) {
		append_tick_to_array(&beep_bank, 0);
	}
	[syncSound unlock];
}
export void _n80_append_sound_ticks(DWORD T)
{
	[syncSound lock];
	beep_bank.current.ticks += (double) T;
	sound_bank.current.ticks += (double) T;
	[syncSound unlock];
}
static void _n80_sound_on(USHORT freq) {
	[syncSound lock];
	if (sound_bank.current.state != freq)
		append_tick_to_array(&sound_bank, freq);
	[syncSound unlock];
}
static void _n80_sound_off() {
	[syncSound lock];
	if (sound_bank.current.state != 0)
		append_tick_to_array(&sound_bank, 0);
	[syncSound unlock];
}
static inline int _n80_make_pcm(UInt32 au_frames, AudioBufferList *ioData)
{
	int16_t		*pcm = ioData->mBuffers[0].mData;
	int 		write_frames = 0, rollover = 0;
	
	[syncSound lock];
	
	//---- Beep ---------------------------------------------------------------
	LPSAMPLETsARR arr1 = &beep_bank.banks[beep_bank.bank];				// current
	LPSAMPLETsARR arr2 = &beep_bank.banks[beep_bank.bank == 0 ? 1 : 0];	// next
	arr2->count = 0;
	arr2->ticks_sum = 0;
	
	// rack of frames
	if (arr1->ticks_sum < au_frames) {
		append_tick_to_array(&beep_bank, beep_bank.current.state);
	}
	for (int i = 0; i < arr1->count; ++i) {
		int frames = (int) floor(arr1->arr[i].ticks + 0.5);
		int remain = frames;
		if (frames > 0) {
			for (int j = 0; j < frames; ++j) {
				// check if one period long over
				if (write_frames >= au_frames) {
					break;
				}
				++write_frames;
				--remain;
				if (arr1->arr[i].state == 0)	{	// beep off
					*pcm++ = 0;
				} else {						// beep on
					if ((n80lx.phase_adjuster % 18) < 9) {
						*pcm++ = WAVETOP;
					} else {
						*pcm++ = WAVEBOTTOM;
					}
				}
				if (++n80lx.phase_adjuster > 1000000)
					n80lx.phase_adjuster = 0;
			}
		}
		if (remain > 0) {
			arr2->arr[rollover].ticks = remain;
			arr2->arr[rollover].state = arr1->arr[i].state;
			arr2->ticks_sum += remain;
			arr2->count++;
			++rollover;
		}
	}
	// bank exchange
	arr1->count     = 0;
	arr1->ticks_sum = 0;
	beep_bank.bank = beep_bank.bank == 0 ? 1 : 0;

	//---- PCG ----------------------------------------------------------------
	pcm = ioData->mBuffers[0].mData;
	write_frames = 0, rollover = 0;
	LPSAMPLETsARR arr10 = &sound_bank.banks[sound_bank.bank];				// current
	LPSAMPLETsARR arr20 = &sound_bank.banks[sound_bank.bank == 0 ? 1 : 0];	// next
	arr20->count = 0;
	arr20->ticks_sum = 0;
	
	// rack of frames
	if (arr10->ticks_sum < au_frames) {
		append_tick_to_array(&sound_bank, sound_bank.current.state);
	}
	for (int i = 0; i < arr10->count; ++i) {
		int frames = (int) floor(arr10->arr[i].ticks + 0.5);
		int remain = frames;
		int wave_frames = arr10->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr10->arr[i].state + 0.5) : 0;
		int half_wave_frames = wave_frames / 2;
		if (frames > 0) {
			for (int j = 0; j < frames; ++j) {
				// check if one period long over
				if (write_frames >= au_frames) {
					break;
				}
				++write_frames;
				--remain;
				if (arr10->arr[i].state > 0)	{	// sound off
					if ((n80lx.phase_adjuster2 % wave_frames) < half_wave_frames) {
						*pcm++ += WAVETOP;
					} else {
						*pcm++ += WAVEBOTTOM;
					}
				}
				if (++n80lx.phase_adjuster2 > 1000000)
					n80lx.phase_adjuster2 = 0;
			}
		}
		if (remain > 0) {
			arr20->arr[rollover].ticks = remain;
			arr20->arr[rollover].state = arr10->arr[i].state;
			arr20->ticks_sum += remain;
			arr20->count++;
			++rollover;
		}
	}
	// bank exchange
	arr10->count     = 0;
	arr10->ticks_sum = 0;
	sound_bank.bank = sound_bank.bank == 0 ? 1 : 0;
	
	[syncSound unlock];
	return write_frames;
}
static OSStatus	audioCallback(
							  void 							*inRefCon,
							  AudioUnitRenderActionFlags 	*ioActionFlags,
							  const AudioTimeStamp 			*inTimeStamp,
							  UInt32 						inBusNumber,
							  UInt32 						inNumberFrames,
							  AudioBufferList				*ioData)

{
	static int	i = 0;
	if (i == 0) {
		i = 1;
		NSLog(@"buffers = %d", ioData->mNumberBuffers);
		NSLog(@"frames = %d, ", inNumberFrames);
		NSLog(@"data size = %d", ioData->mBuffers[0].mDataByteSize);
		NSLog(@"channels = %d", ioData->mBuffers[0].mNumberChannels);
	}
	memset(ioData->mBuffers[0].mData, 0, ioData->mBuffers[0].mDataByteSize);
	_n80_make_pcm(inNumberFrames, ioData);
	return noErr;
}
static inline double getclock() {
	uint64_t abstm = mach_absolute_time();
    static double conversion = 0.0;
    
    if (conversion == 0.0) {
		mach_timebase_info_data_t info;
		kern_return_t err = mach_timebase_info(&info);
		//Convert the timebase into seconds
		if (err == 0)
			conversion = 1e-9 * (double) info.numer / (double) info.denom;
    }
    return conversion * (double) abstm;
}
//-----------------------------------------------------------------------------
//	Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80em
- (void)dealloc
{
	NSLog(@"n80em dealloc");
	[self reqDispose];
    [super dealloc];
}
- (BOOL)initEmulator:(NSView *)myView {
	// init member variables
	bitmap = NULL;
	bitmap1 = NULL;
	bitmap2 = NULL;
	bitmapbank = 0;
	bmpctx = NULL;
	gOutputUnit = NULL;
	Z80running  = false;
	Disprunning = false;
	
	// N80 structure clean
	memset(&n80lx, 0, sizeof(n80lx));
	
	// Z80 Initialize
	__z80_set_external_method(&n80lx.cpu, _n80_mbread, _n80_mbwrite, _n80_ioread, _n80_iowrite);
	
	// PC-8001 Device Initialize
	_n80_device_initialize(&n80device, enter, leave, chg_bltfunc);
	n80device.beep_on = _n80_beep_on;
	n80device.beep_off = _n80_beep_off;
	n80device.sound_on = _n80_sound_on;
	n80device.sound_off = _n80_sound_off;
	
	// beep/sound bank init
	memset(&beep_bank, 0, sizeof(beep_bank));
	memset(&sound_bank, 0, sizeof(sound_bank));
	beep_bank.current.state = 1;
	beep_bank.current.ticks = n80device.param.sync_ticks * 4;
	sound_bank.current.state = 0;
	sound_bank.current.ticks = 0;
	// CPU Ticks変換係数: frames = (44100/3993600*Ts)
	n80lx.sampling_rate = 44100;
	n80lx.frame_keisu = (double) n80lx.sampling_rate / (double) n80device.param.hz;
	
	// load font file
	if (![self loadFontFile])
		return false;
	
	// load rom file
	if (![self loadROMFile])
		return false;
	
	// offscreen bitmap
	bitmap1 = calloc(640*400, 4);
	if ((void *) bitmap1 == NULL) {
		NSLog(@"Memory not allocated!");
		return false;
	}
	bitmap2 = calloc(640*400, 4);
	if ((void *) bitmap2 == NULL) {
		NSLog(@"Memory not allocated!");
		free(bitmap1);
		bitmap1 = NULL;
		return false;
	}
	bitmap = bitmap1;
	bitmapbank = 0;
	[self prepareDotPatterns];
	if (![self MyCreateBitmapContext:640 :400])
		return false;
	
	if (![self prepareAudio])
		return false;
	
	// save NSView
	view = myView;
	return true;
}
- (void)startEmulator {
	joinZ80  = [[NSConditionLock alloc] initWithCondition:NOTDONE];
//	joinDisp = [[NSConditionLock alloc] initWithCondition:NOTDONE];
	pauseZ80sync = [[NSConditionLock alloc] initWithCondition:NOTDONE];
	syncSound = [[NSConditionLock alloc] initWithCondition:NOTDONE];
	[NSThread detachNewThreadSelector:@selector(z80Thread)
							 toTarget:self
						   withObject:nil];
//	[NSThread detachNewThreadSelector:@selector(dispThread)
//							 toTarget:self
//						   withObject:nil];
	Z80running  = true;
	Disprunning = true;
}
- (void)reqDispose {
	// stop sound unit
	if (gOutputUnit != NULL) {
		AudioOutputUnitStop(gOutputUnit);
		AudioUnitUninitialize(gOutputUnit);
		AudioComponentInstanceDispose(gOutputUnit);
	}
	// stop Z80/display thread
	OSAtomicCompareAndSwap32(0, 1, &n80lx.request_cpu_signal);
	OSAtomicCompareAndSwap32(0, 1, &n80lx.request_disp_signal);
	if (Z80running) {
		Z80running = false;
		[joinZ80 lockWhenCondition:DONE];
		[joinZ80 unlock];
		[joinZ80 release];
		[pauseZ80sync release];
	}
//	if (Disprunning) {
//		Disprunning = false;
//		[joinDisp lockWhenCondition:DONE];
//		[joinDisp unlock];
//		[joinDisp release];
//	}
	if (syncSound != NULL) {
		[syncSound release];
		syncSound = NULL;
	}
	// release resources
	if (bmpctx != NULL) {
		CGContextRelease(bmpctx);
		bmpctx = NULL;
	}
	if (bitmap1 != NULL) {
		free(bitmap1);
		bitmap1 = NULL;
	}
	if (bitmap2 != NULL) {
		free(bitmap2);
		bitmap2 = NULL;
	}
	view = NULL;
}
- (void)reqReset {
	OSAtomicCompareAndSwap32(0, 2, &n80lx.request_cpu_signal);
}
- (void)reqPCGToggle {
	n80device.pcg_mode ^= 0x01;
	n80device.attr.invalid();
}
- (void)reqOpenFile {
	[pauseZ80sync lock];
	OSAtomicCompareAndSwap32(0, 3, &n80lx.request_cpu_signal);
	OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);

	FILE *fp = NULL;
	BOOL isN80File = false;
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"N80",@"'N80 Memory Image File'",@"CMT",@"'Casset Image File'",nil];
	[openPanel setAllowedFileTypes:allowedFileTypes];
	NSInteger pressedButton = [openPanel runModal];
	if (pressedButton == NSOKButton) {
		NSString *path = [[openPanel URL] path];
		isN80File = ([path hasSuffix:@"N80"] || [path hasSuffix:@"n80"]);
		fp = fopen([path fileSystemRepresentation],"r");
	}
	if (fp != NULL) {
		if (isN80File) {	// N80 File
			USHORT spval;
			
			fseek(fp, 0, SEEK_END);
			long fsiz = ftell(fp);
			rewind(fp);
			if (fsiz >= STDSIZE) {
				fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
				fread(n80device.mem.map._ram, 1, fsiz, fp);
				fseek(fp, SP_VAL, SEEK_SET);
				fread(&spval, 2, 1, fp);
				n80lx.cpu.r.x.PC = PC_VAL;
				n80lx.cpu.r.x.SP = spval;
			}
		} else {			// CMT File
			size_t len = fread(n80lx.cmt_buff, 1, (1024*32), fp);
			n80device.cmt._size  = (int) len;
			n80device.cmt._buff  = n80lx.cmt_buff;
			n80device.cmt._index = 0;
		}
		fclose(fp);
	}
	[pauseZ80sync unlock];
}
- (CGImageRef)reqImage {
	if (bmpctx != NULL)
		return CGBitmapContextCreateImage(bmpctx);
	else
		return NULL;
}
- (void *)reqTexBitmap {
	return bitmapbank == 0 ? bitmap2 : bitmap1;
	/*if (bitmap != NULL)
		return bitmap;
	else
		return NULL;*/
}
//-----------------------------------------------------------------------------
//	Initialize Emulator
//-----------------------------------------------------------------------------
- (BOOL)loadFontFile {
	// find font file from App local path
	NSString *strPath = [[NSBundle mainBundle] resourcePath];
	NSString *defPath = [strPath stringByAppendingString:@"/pc-8001.fon"];
	
	FILE *fp = NULL;
	if ((fp = fopen([defPath fileSystemRepresentation],"r")) == NULL) {
		// choose font file by user
		NSOpenPanel *openPanel = [NSOpenPanel openPanel];
		NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"fon",@"'FONT'",nil];
		[openPanel setAllowedFileTypes:allowedFileTypes];
		NSInteger pressedButton = [openPanel runModal];
		if (pressedButton == NSOKButton) {
			NSString *path = [[openPanel URL] path];
			fp = fopen([path fileSystemRepresentation],"r");
		}
	}
	if (fp == NULL)
		return false;
	
	// read font data from font file
	size_t len = fread(&n80device.font, 1, (1024*4), fp);
	fclose(fp);
	if (len != (1024*4))
		return false;
	
	// copy font rom data to PCG ram
	for (int i = 0; i < 2048; ++i)
		n80device.pcg[i] = n80device.font[i];
	
	return true;
}
- (BOOL)loadROMFile {
	// find rom file from App local path
	NSString *strPath = [[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent];
	NSString *defPath = [strPath stringByAppendingString:@"/pc-8001.rom"];
	
	FILE *fp = NULL;
	if ((fp = fopen([defPath fileSystemRepresentation],"r")) == NULL) {
		NSString *defPath2 = [strPath stringByAppendingString:@"/8801-n80.rom"];
		if ((fp = fopen([defPath2 fileSystemRepresentation],"r")) == NULL) {
			// choose rom file by user
			NSOpenPanel *openPanel = [NSOpenPanel openPanel];
			NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"rom",@"'ROM'",nil];
			[openPanel setAllowedFileTypes:allowedFileTypes];
			NSInteger pressedButton = [openPanel runModal];
			if (pressedButton == NSOKButton) {
				NSString *path = [[openPanel URL] path];
				fp = fopen([path fileSystemRepresentation],"r");
			}
		}
	}
	if (fp == NULL)
		return false;
	
	// read rom data from font file
	size_t len = fread(&n80device.mem.map._rom[0], 1, (1024*32), fp);
	fclose(fp);
	if (len != (1024*32) && len != (1024*24))
		return false;
	return true;
}
- (void)prepareDotPatterns {
	// fb用ビットパターンデータを作成
	for (int i = 0; i < 8; ++i) {
		DWORD color = SPITZ_COLOR[i];
		for (int j = 0; j < 256; ++j) {
			BYTE bit = (BYTE) (j & 0xff);
			n80lx.dot_wpattern[i][j][0]  = bit & 0x80 ? color : 0;
			n80lx.dot_wpattern[i][j][1]  = bit & 0x80 ? color : 0;
			n80lx.dot_wpattern[i][j][2]  = bit & 0x40 ? color : 0;
			n80lx.dot_wpattern[i][j][3]  = bit & 0x40 ? color : 0;
			n80lx.dot_wpattern[i][j][4]  = bit & 0x20 ? color : 0;
			n80lx.dot_wpattern[i][j][5]  = bit & 0x20 ? color : 0;
			n80lx.dot_wpattern[i][j][6]  = bit & 0x10 ? color : 0;
			n80lx.dot_wpattern[i][j][7]  = bit & 0x10 ? color : 0;
			n80lx.dot_wpattern[i][j][8]  = bit & 0x08 ? color : 0;
			n80lx.dot_wpattern[i][j][9]  = bit & 0x08 ? color : 0;
			n80lx.dot_wpattern[i][j][10] = bit & 0x04 ? color : 0;
			n80lx.dot_wpattern[i][j][11] = bit & 0x04 ? color : 0;
			n80lx.dot_wpattern[i][j][12] = bit & 0x02 ? color : 0;
			n80lx.dot_wpattern[i][j][13] = bit & 0x02 ? color : 0;
			n80lx.dot_wpattern[i][j][14] = bit & 0x01 ? color : 0;
			n80lx.dot_wpattern[i][j][15] = bit & 0x01 ? color : 0;
			
			n80lx.dot_npattern[i][j][0]  = bit & 0x80 ? color : 0;
			n80lx.dot_npattern[i][j][1]  = bit & 0x40 ? color : 0;
			n80lx.dot_npattern[i][j][2]  = bit & 0x20 ? color : 0;
			n80lx.dot_npattern[i][j][3]  = bit & 0x10 ? color : 0;
			n80lx.dot_npattern[i][j][4]  = bit & 0x08 ? color : 0;
			n80lx.dot_npattern[i][j][5]  = bit & 0x04 ? color : 0;
			n80lx.dot_npattern[i][j][6]  = bit & 0x02 ? color : 0;
			n80lx.dot_npattern[i][j][7]  = bit & 0x01 ? color : 0;
		}
	}
}
- (BOOL) MyCreateBitmapContext:(int) pixelsWide : (int) pixelsHigh {
    CGColorSpaceRef	colorSpace;
	int				bitmapByteCount;
	int				bitmapBytesPerRow;
	
	bitmapBytesPerRow   = (pixelsWide * 4);
	bitmapByteCount     = (bitmapBytesPerRow * pixelsHigh);
	colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
	bmpctx = CGBitmapContextCreate (bitmap,
									pixelsWide,
									pixelsHigh,
									8,	// bits per component (r/g/b)
									bitmapBytesPerRow,
									colorSpace,
									kCGImageAlphaNoneSkipLast);
    if (bmpctx == NULL) {
		free (bitmap);
		bitmap = NULL;
		NSLog(@"Context not created!");
		return false;
    }
	return true;
}
- (BOOL)prepareAudio {
	OSStatus	err = noErr;
	int			sampleRate = SAMPLE_RATE;
	
	// Open the default output unit
	AudioComponentDescription desc;
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	
	AudioComponent comp = AudioComponentFindNext(NULL, &desc);
	if (comp == NULL) {
		NSLog(@"AudioComponentFindNext error");
		return false;
	}
	err = AudioComponentInstanceNew(comp, &gOutputUnit);
	if (comp == NULL) {
		NSLog(@"AudioComponentInstanceNew=%d", err);
		return false;
	}
	err = AudioUnitInitialize(gOutputUnit);
	
	// Set up a callback function to generate output to the output unit
    AURenderCallbackStruct callback;
	callback.inputProc = audioCallback;
	callback.inputProcRefCon = self;
	
	err = AudioUnitSetProperty (gOutputUnit,
								kAudioUnitProperty_SetRenderCallback,
								//kAudioOutputUnitProperty_EnableIO,
								kAudioUnitScope_Input,		//initial value is "kAudioUnitScope_Input". is it ok?
								0,
								&callback,
								sizeof(callback));
	
	AudioStreamBasicDescription streamFormat;
	streamFormat.mSampleRate = sampleRate;					//	the sample rate of the audio stream
	streamFormat.mFormatID = kAudioFormatLinearPCM;			//	the specific encoding type of audio stream
#ifdef __BIG_ENDIAN__
	streamFormat.mFormatFlags =		kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsBigEndian
	| kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;		//	16bit int
#else
	streamFormat.mFormatFlags =		kLinearPCMFormatFlagIsSignedInteger
	| kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;		//	16bit int
#endif
	streamFormat.mBytesPerPacket = 2;
	streamFormat.mFramesPerPacket = 1;
	streamFormat.mBytesPerFrame = 2;
	streamFormat.mChannelsPerFrame = 1;
	streamFormat.mBitsPerChannel = 16;
	
	err = AudioUnitSetProperty (gOutputUnit,
								kAudioUnitProperty_StreamFormat,
								kAudioUnitScope_Input,
								0,
								&streamFormat,
								sizeof(AudioStreamBasicDescription));
		
	Float64 outSampleRate;
	UInt32 size = sizeof(Float64);
	
	err = AudioUnitGetProperty (gOutputUnit,
								kAudioUnitProperty_SampleRate,
								kAudioUnitScope_Output,
								0,
								&outSampleRate,
								&size);
	
	err = AudioOutputUnitStart (gOutputUnit);
	
	return true;
}
//-----------------------------------------------------------------------------
//	Z80 CPU Thread
//-----------------------------------------------------------------------------
- (void)z80Thread {
	
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
//	NSOperationQueue *queue = [[[NSOperationQueue alloc] init] autorelease];
	[joinZ80 lock];
	@try {
		volatile double clock_sync = getclock() + 0.02;
		volatile DWORD state = 0, dma_penalty = 0;
		volatile int quit = 0;
#ifdef DEBUG
//		volatile double lastdps = getclock(), curdps;
//		volatile int dps_counter = 0;
		int wait_total = 0;
		int wait_sleep = 0;
#endif
		static int32_t sync = 0;
		self.pause = false;

		for (; !quit;) {
			if (OSAtomicCompareAndSwap32(1, 0, &n80lx.request_cpu_signal)) {
				NSLog(@"Accept Request CPU HALT");
				break;
				
			} else if (OSAtomicCompareAndSwap32(2, 0, &n80lx.request_cpu_signal)) {
				NSLog(@"Accept Request CPU RESET");
				n80lx.cpu.reset(&n80lx.cpu);
				n80device.attr.invalid();
				n80device.clear_keys(&n80device);
				[syncSound lock];					// reset sound tick buffer
				beep_bank.bank = 0;
				beep_bank.banks[0].count = 0;
				beep_bank.banks[0].ticks_sum = 0;
				beep_bank.current.state = 0;
				beep_bank.current.ticks = 0;
				sound_bank.bank = 0;
				sound_bank.banks[0].count = 0;
				sound_bank.banks[0].ticks_sum = 0;
				sound_bank.current.state = 0;
				sound_bank.current.ticks = 0;
				[syncSound unlock];
				clock_sync = getclock() + 0.02;
				state       = 0;
				dma_penalty = 0;
				
			} else if (OSAtomicCompareAndSwap32(3, 0, &n80lx.request_cpu_signal)) {
				NSLog(@"Accept Request CPU PAUSE");
				self.pause = true;
				[pauseZ80sync lock];
				[pauseZ80sync unlock];
				self.pause = false;
				clock_sync = getclock() + 0.02;
				state       = 0;
				dma_penalty = 0;
			}
			if (n80lx.cpu.r.x.HALT != 0)
				continue;
			
			// z80 step execute
			if (dma_penalty > 0) {
				if (dma_penalty > 20) {
					n80lx.cpu.r.x.T  = 20;
					dma_penalty     -= 20;
				} else {
					n80lx.cpu.r.x.T  = dma_penalty;
					dma_penalty      = 0;
				}
				n80lx.cpu.r.x.W = 0;
			} else {
				if (n80lx.cpu.step(&n80lx.cpu)) {
					dispatch_async(dispatch_get_main_queue(), ^{
						NSAlert *alert = [[NSAlert alloc] init];
						[alert addButtonWithTitle:@"OK"];
						[alert setMessageText:@"detect HALT or illegal instruction"];
						[alert setInformativeText:@"reset a emulator automatically."];
						[alert setAlertStyle:NSWarningAlertStyle];
						[alert runModal];
						[alert release];
						[self reqReset];
					});
				}
			}
			n80lx.cpu.r.x.T += n80lx.cpu.r.x.W;
			
			// vrtc operation
			if (n80device.vrtc.vsync_line < 200) {
				n80device.vrtc.hsync_state += n80lx.cpu.r.x.T;
				if (n80device.vrtc.hsync_state < 179)
					;
				else if (n80device.vrtc.hsync_state < 250)
					;
				else {
					dma_penalty = 70;
					if (++n80device.vrtc.vsync_line >= 200) {
						n80device.vrtc.vsync_state = 0;
						n80device.vrtc._vrtc_flag  = 0xff;
						n80device.vrtc.edge();
#ifdef DEBUG			// dps calc
/*						++dps_counter;
						curdps = getclock();
						double tick = curdps - lastdps;
						if (tick > 1.0) {
							lastdps = curdps;
							double dps = (double) dps_counter / tick;
							dps_counter = 0;
							NSLog(@"%f dps", dps);
						}*/
#endif					// display operation launch via working thread
						//[queue addOperationWithBlock:^{
#ifdef DEBUG
							//if ([NSThread isMainThread])
							//	NSLog(@"NSOperation start with main thread");
#endif						// check if operation is overlapped
							//if (OSAtomicCompareAndSwap32(0, 1, &sync)) {
								[self refleshScreen];
								if (bitmapbank == 0) {
									bitmap = bitmap2;
									bitmapbank = 1;
								} else {
									bitmap = bitmap1;
									bitmapbank = 0;
								}
								dispatch_async(dispatch_get_main_queue(), ^{
									[view setNeedsDisplay:TRUE];
								});
								sync = 0;
							//}
#ifdef DEBUG
							//else {
							//	NSLog(@"NSOperation is overlapped");
							//}
#endif
						//}];
					} else
						n80device.vrtc.hsync_state -= 250;
				}
			} else {
				n80device.vrtc.vsync_state += n80lx.cpu.r.x.T;
				if (n80device.vrtc.vsync_state > 14016) {
					n80device.vrtc.vsync_line   = 0;
					n80device.vrtc.hsync_state  = n80device.vrtc.vsync_state - 14016;
					n80device.vrtc._vrtc_flag   = 0;
				}
			}
			// beep/pcg ticks
			_n80_append_sound_ticks(n80lx.cpu.r.x.T);
			
			// cpu clock sync
			state += n80lx.cpu.r.x.T;
			if (state >= n80device.param.sync_ticks) {
#if DEBUG
				++wait_total;
#endif
				double present = getclock();
				if (present < clock_sync) {
					double diff = clock_sync - present;
					if (diff >= 0.005) {	// 5ms以上ある場合はsleep
#if DEBUG
						++wait_sleep;
#endif
						//NSLog(@"sleeping %f", (float) diff);
						DWORD  slp  = (DWORD) (diff * 1000000);
						usleep(slp);
					}
					state -= n80device.param.sync_ticks;
				} else
					state = 0;
				clock_sync += 0.02;	// 20ms
				if ((present = getclock()) >= clock_sync)
					clock_sync = present + 0.02;
				//			while (getclock() >= clock_sync)
				//				clock_sync += 0.02;
#if DEBUG
				if (wait_total > 100) {
					float percent = ((float)wait_sleep / (float)wait_total) * 100;
					printf("wait sleep rate = %f%%\n", percent);
					wait_total = 0;
					wait_sleep = 0;
				}
#endif
			}
		}
	}
	@catch (NSException *exception) {
		NSLog(@"z80Thread caught exception %@", [exception reason]);
	}
	@finally {
		[joinZ80 unlockWithCondition:DONE];
		[pool release];
	}
}
//-----------------------------------------------------------------------------
//	Display Thread
//-----------------------------------------------------------------------------
#if DEBUG
static int show_eval = 0;
#endif
- (void)refleshScreen {
	int i, change_attr = 0, change_prev = 0;

	for (i = 0; i < 25; ++i) {
		n80device.vram_sync_enter();
		/*if (n80device.attr._attrchg[i] == 1 || change_prev) {
			n80device.attr.cache(i);
			change_attr = 1;
		} else
			change_attr = 0;*/
		n80device.attr.cache(i);
		change_attr = 1;
		n80device.vram_sync_leave();
		if (change_attr) {
			change_prev = n80device.attr.eval(i) == 1 ? 1 : 0;
#if DEBUG
			if (show_eval != 0) {
				printf("row = %2d: ", i);
				for (int c = 0; c < 10; ++c) {
					printf("%2.2x ", n80device.attr._eval_attr[c]);
				}
				printf("\n");
			}
#endif
			switch (bltmode) {
				case 0:
					[self blt4020:i];
					break;
				case 1:
					[self blt4025:i];
					break;
				case 2:
					[self blt8020:i];
					break;
				default:
					[self blt8025:i];
					break;
			}
		}
	}
}
- (void)blt8025:(int) line {
	if (line < 25) {
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		
		for (int col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			for (int r = 0; r < 8; ++r) {
				int		 bmpidx = col * 8 + (640 * (399 - (line * 16 + r * 2)));
				BYTE	 bit = *fnt++;
				LPDWORD	 ptr;
				
				if (attr & 0x40)			// secret
					bit = 0;
				if (attr & 0x80 && r == 7)	// under line
					bit |= 0xff;
				if (attr & 0x08 && r == 0)	// upper line
					bit |= 0xff;
				if (attr & 0x20)			// reverse
					bit = ~bit;
				
				ptr = &n80lx.dot_npattern[color][bit][0];
				for (int h = 0; h < 8; ++h) {
					DWORD pix = *ptr++;
					bitmap[bmpidx]       = pix;
					bitmap[bmpidx - 640] = pix;
					++bmpidx;
				}
			}
		}
	}
}
- (void)blt8020:(int) line {
	if (line < 20) {
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		
		for (int col = 0; col < 80; ++col) {
			BYTE		ascii = *txts++;
			BYTE		attr  = *pacs++;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			for (int r = 0; r < 10; ++r) {
				int		 bmpidx = col * 8 + (640 * (399 - (line * 20 + r * 2)));
				BYTE	 bit;
				LPDWORD	 ptr;
				
				if (r > 0 && r < 9)
					bit = *fnt++;
				else
					bit = 0;
				
				if (attr & 0x40)			// secret
					bit = 0;
				if (attr & 0x80 && r == 9)	// under line
					bit |= 0xff;
				if (attr & 0x08 && r == 0)	// upper line
					bit |= 0xff;
				if (attr & 0x20)			// reverse
					bit = ~bit;
				
				ptr = &n80lx.dot_npattern[color][bit][0];
				for (int h = 0; h < 8; ++h) {
					DWORD pix = *ptr++;
					bitmap[bmpidx]       = pix;
					bitmap[bmpidx - 640] = pix;
					++bmpidx;
				}
			}
		}
	}
}
- (void)blt4025:(int) line {
	if (line < 25) {
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		
		for (int col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			
			txts += 2;
			pacs += 2;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			for (int r = 0; r < 8; ++r) {
				int		 bmpidx = col * 8 + (640 * (399 - (line * 16 + r * 2)));
				BYTE	 bit = *fnt++;
				LPDWORD	 ptr;
				
				if (attr & 0x40)			// secret
					bit = 0;
				if (attr & 0x80 && r == 7)	// under line
					bit |= 0xff;
				if (attr & 0x08 && r == 0)	// upper line
					bit |= 0xff;
				if (attr & 0x20)			// reverse
					bit = ~bit;
				
				ptr = &n80lx.dot_wpattern[color][bit][0];
				for (int h = 0; h < 16; ++h) {
					DWORD pix = *ptr++;
					bitmap[bmpidx]       = pix;
					bitmap[bmpidx - 640] = pix;
					++bmpidx;
				}
			}
		}
	}
}
- (void)blt4020:(int) line {
	if (line < 20) {
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		
		for (int col = 0; col < 80; col += 2) {
			BYTE		ascii = *txts;
			BYTE		attr  = *pacs;
			LPBYTE		fnt   = &n80device.font[ascii*8];
			int		 	color = attr & 0x07;
			
			txts += 2;
			pacs += 2;
			
			if (attr & 0x10)				// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			for (int r = 0; r < 10; ++r) {
				int		 bmpidx = col * 8 + (640 * (399 - (line * 20 + r * 2)));
				BYTE	 bit;
				LPDWORD	 ptr;
				
				if (r > 0 && r < 9)
					bit = *fnt++;
				else
					bit = 0;

				if (attr & 0x40)			// secret
					bit = 0;
				if (attr & 0x80 && r == 9)	// under line
					bit |= 0xff;
				if (attr & 0x08 && r == 0)	// upper line
					bit |= 0xff;
				if (attr & 0x20)			// reverse
					bit = ~bit;
				
				ptr = &n80lx.dot_wpattern[color][bit][0];
				for (int h = 0; h < 16; ++h) {
					DWORD pix = *ptr++;
					bitmap[bmpidx]       = pix;
					bitmap[bmpidx - 640] = pix;
					++bmpidx;
				}
			}
		}
	}
}
//-----------------------------------------------------------------------------
//	Key Event Handler
//-----------------------------------------------------------------------------
- (void)OnKeyDown:(NSEvent *) evt {
	unichar ch = [[evt characters] characterAtIndex:0];
	//NSLog(@"KeyDown %d",ch);
	BYTE	key = [self convrtykey:ch];
	[self sendN80KeyDown:key];
#ifdef DEBUG
	if (ch == 94) {
		show_eval = 1;
	}
#endif
}
- (void)OnKeyUp:(NSEvent *) evt {
	unichar ch = [[evt characters] characterAtIndex:0];
	//NSLog(@"KeyUp %d",ch);
	BYTE	key = [self convrtykey:ch];
	[self sendN80KeyUp:key];
#ifdef DEBUG
	if (ch == 94) {
		[self print_attribute];
		show_eval = 0;
	}
#endif
}
- (void)OnFlagsChanged: (NSEvent *) evt {
	NSUInteger modifierFlags = [[NSApp currentEvent] modifierFlags];
	if (modifierFlags & NSShiftKeyMask) {		//SHIFT
		[self sendN80KeyDown:0x46];
	} else {
		[self sendN80KeyUp:0x46];
	}
	if (modifierFlags & NSControlKeyMask) {		//CTL
		[self sendN80KeyDown:0x47];
	} else {
		[self sendN80KeyUp:0x47];
	}
	if (modifierFlags & NSAlternateKeyMask) {	//GRPH
		[self sendN80KeyDown:0x44];
	} else {
		[self sendN80KeyUp:0x44];
	}
	if (modifierFlags & NSCommandKeyMask) {		//KANA
		[self sendN80KeyDown:0x45];
	} else {
		[self sendN80KeyUp:0x45];
	}
	self->fnKey = (modifierFlags & NSFunctionKeyMask) != 0;	//fn key
	if (self->fnKey)
		printf("fn key on!\n");
	else
		printf("fn key off\n");
}
- (void) sendN80KeyDown:(BYTE)key {
	if (key != 0xff) {
		int  index = down_matorix[key][0];
		BYTE bit   = down_matorix[key][1];
		n80device.keys[index] &= bit;
	}
}
- (void) sendN80KeyUp:(BYTE)key {
	if (key != 0xff) {
		int  index = up_matorix[key][0];
		BYTE bit   = up_matorix[key][1];
		n80device.keys[index] |= bit;
	}
}
- (BYTE)convrtykey:(unichar) code {
	BYTE	key = 0xff;
	switch (code) {
		case 63236:	/* F1 */	key = 0x49;	break;
		case 63237:	/* F2 */	key = 0x4A;	break;
		case 63238:	/* F3 */	key = 0x4B;	break;
		case 63239:	/* F4 */	key = 0x4C;	break;
		case 63240:	/* F5 */	key = 0x4D;	break;
			
		case 33:	/* 1 */	key = 0x31;	break;
		case 34:	/* 2 */	key = 0x32;	break;
		case 35:	/* 3 */	key = 0x33;	break;
		case 36:	/* 4 */	key = 0x34;	break;
		case 37:	/* 5 */	key = 0x35;	break;
		case 38:	/* 6 */	key = 0x36;	break;
		case 39:	/* 7 */	key = 0x37;	break;
		case 40:	/* 8 */	key = 0x38;	break;
		case 41:	/* 9 */	key = 0x39;	break;

		case 49:	/* 1 */	key = 0x31;	break;
		case 50:	/* 2 */	key = 0x32;	break;
		case 51:	/* 3 */	key = 0x33;	break;
		case 52:	/* 4 */	key = 0x34;	break;
		case 53:	/* 5 */	key = 0x35;	break;
		case 54:	/* 6 */	key = 0x36;	break;
		case 55:	/* 7 */	key = 0x37;	break;
		case 56:	/* 8 */	key = 0x38;	break;
		case 57:	/* 9 */	key = 0x39;	break;
		case 48:	/* 0 */	key = 0x30;	break;
			
		case 0x51:
		case 0x71:	/* Q */	key = 0x21;	break;
		case 0x57:
		case 0x77:	/* W */	key = 0x27;	break;
		case 0x45:
		case 0x65:	/* E */	key = 0x15;	break;
		case 0x52:
		case 0x72:	/* R */	key = 0x22;	break;
		case 0x54:
		case 0x74:	/* T */	key = 0x24;	break;
		case 0x59:
		case 0x79:	/* Y */	key = 0x29;	break;
		case 0x55:
		case 0x75:	/* U */	key = 0x25;	break;
		case 0x49:
		case 0x69:	/* I */	key = 0x19;	break;
		case 0x4f:
		case 0x6f:	/* O */	key = 0x1F;	break;
		case 0x50:
		case 0x70:	/* P */	key = 0x20;	break;

		case 0x41:
		case 0x61:	/* A */	key = 0x11;	break;
		case 0x53:
		case 0x73:	/* S */	key = 0x23;	break;
		case 0x44:
		case 0x64:	/* D */	key = 0x14;	break;
		case 0x46:
		case 0x66:	/* F */	key = 0x16;	break;
		case 0x47:
		case 0x67:	/* G */	key = 0x17;	break;
		case 0x48:
		case 0x68:	/* H */	key = 0x18;	break;
		case 0x4a:
		case 0x6a:	/* J */	key = 0x1A;	break;
		case 0x4b:
		case 0x6b:	/* K */	key = 0x1B;	break;
		case 0x4c:
		case 0x6c:	/* L */	key = 0x1C;	break;
			
		case 0x5a:
		case 0x7a:	/* Z */	key = 0x2A;	break;
		case 0x58:
		case 0x78:	/* X */	key = 0x28;	break;
		case 0x43:
		case 0x63:	/* C */	key = 0x13;	break;
		case 0x56:
		case 0x76:	/* V */	key = 0x26;	break;
		case 0x42:
		case 0x62:	/* B */	key = 0x12;	break;
		case 0x4e:
		case 0x6e:	/* N */	key = 0x1E;	break;
		case 0x4d:
		case 0x6d:	/* M */	key = 0x1D;	break;
			
		case 0x2d:
		case 0x3d:	/* =     */	key = 0x2F;	break;
		case 0x2b:
		case 0x3b:	/* ;     */	key = 0x3B;	break;
		case 0x2a:
		case 0x3a:	/* :     */	key = 0x3A;	break;
		case 0x5e:
		case 0x7e:	/* ^     */	key = 0x2E;	break;
		case 0x40:
		case 0x60:	/* @     */	key = 0x10;	break;
		case 0x5f:	/* _     */	key = 0x3F;	break;
		case 0x5b:
		case 0x7b:	/* [     */	key = 0x2B;	break;
		case 0x5d:
		case 0x7d:	/* ]     */	key = 0x2D;	break;
		case 0xa5:
		case 0x5c:
		case 0x7c:	/* ¥     */	key = 0x2C;	break;
			
		case 127:	/* BS    */	key = 0x43;	break;
//		case   9:	/* TAB	 */	key = 0x50;	break;
		case  13:	/* ENTER */	key = 0x0F;	break;
//		case 0x1D:	/* CTL   */	key = 0x47;	break;
//		case 0x2A:	/* L-SHF */	key = 0x46;	break;
		case  44:	/* < ,   */	key = 0x3C;	break;
		case  47:	/* /     */	key = 0x3E;	break;
		case  46:	/* > .   */	key = 0x3D;	break;
//		case 0x36:	/* R-SHF */	key = 0x46;	break;
//		case  104:	/* KANA  */	key = 0x45;	break;
		case  32:	/* SP    */	key = self->fnKey ? 0x45 : 0x4E;	break;
//		case 0x3E:	/* CAN   */	key = n80lx.fn_on ? 0x48 : 0x4F;	break;	// STOP/ESC
		case  27:	/* ESC   */	key = self->fnKey ? 0x40 : 0x4F;	break;	// HOME(+fn)/ESC
//		case 0x61:	/* ZEN   */	key = 0x44;	break;	// GRAPH
//		case 0x63:	/* HOME  */	key = 0x40;	break;
			
//		case 0x67:	/* ARW-U */	key = n80lx.fn_on ? 0x41 : 0x08;	break;
//		case 0x69:	/* ARW-L */	key = n80lx.fn_on ? 0x42 : 0x04;	break;
//		case 0x6A:	/* ARW-R */	key = n80lx.fn_on ? 0x41 : 0x06;	break;
//		case 0x6C:	/* ARR-D */	key = n80lx.fn_on ? 0x42 : 0x02;	break;
		case 63232:	/* ARW-U */	key = 0x08;	break;
		case 63234:	/* ARW-L */	key = 0x04;	break;
		case 63235:	/* ARW-R */	key = 0x06;	break;
		case 63233:	/* ARR-D */	key = 0x02;	break;
		case 63276:	/* ARW-U */	key = 0x41;	break;	//+fn
		case 63273:	/* ARW-L */	key = 0x42;	break;
		case 63275:	/* ARW-R */	key = 0x42;	break;
		case 63277:	/* ARR-D */	key = 0x41;	break;
		case 63272:	/* BS */	key = 0x48;	break;	//+fn STOP
			
		default:	break;
	}
#ifdef DEBUG
//	NSLog(@"convrtykey %d -> %d",code, key);
#endif
	return key;
}
#ifdef DEBUG
-(void)print_attribute {
	int line = 3;
	
	for (line = 0; line < 25; ++line) {
		LPBYTE attr = &n80device.mem.arr[0xf300 + 80 + (120*line)];
		printf("row = %d\n", line);
		for (int col = 0; col < 10; ++col) {
			printf("%2.2x-%2.2x ", attr[col*2], attr[col*2+1]);
		}
		printf("\n");
		for (int col = 10; col < 20; ++col) {
			printf("%2.2x-%2.2x ", attr[col*2], attr[col*2+1]);
		}
		printf("\n\n");
	}
}
#endif
@end

