/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "pc-8001.h"
#include <memory.h>
#include <math.h>
#ifdef __cplusplus
extern "C"
{
#endif
#define export
extern void _n80_chg_converter(int mode);			// ATTRIB.c
extern void _n80_attr_invalid();
extern void _n80_attr_set(int line, BYTE flg);
extern void _n80_attr_cache(int line);
extern int  _n80_eval_attrib(int line);
extern void _n80_init_calender();					// PC1990.c
extern void _n80_time_mode();
extern void _n80_width_mode(BYTE v);				// CRTC.c
extern void _n80_crtc_cmd(BYTE v);
extern void _n80_crtc_data(BYTE v);
extern BYTE _n80_poll_vrtc(BYTE i40);				// VRTC.c
extern void _n80_renew_vrtc();
extern void _n80_cursor_enable();					// CURSOR.c
extern void _n80_cursor_disable();
extern void _n80_move_cursor();
extern BYTE _n80_cmt_read();						// CMT.c
extern void _n80_cmt_command(BYTE v);
extern void _n80_pcg_mode(BYTE portno, BYTE data);	// PCG.c
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
 N80DEVICE	n80device;
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
static void clear_io(LPN80DEVICE dev)
{
	memset(&dev->io, 0, sizeof(N80IOPORT));
}

static void clear_ram(LPN80DEVICE dev)
{
	LPBYTE	mem = dev->mem.map._ram;
	int  i;

    for (i = 0; i < 0x1000; ++ i) {
		*mem++ = 0x00;
		*mem++ = 0xff;
	}
	for (i = 0; i < 0x1000; ++ i) {
		*mem++ = 0xff;
		*mem++ = 0x00;
	}
	for (i = 0; i < 128; ++ i) {
		int  j;

		for (j = 0; j < 64; ++j)
		    *mem++ = 0xff;
		for (j = 0; j < 64; ++j)
		    *mem++ = 0x00;
	}
}

static void clear_keys(LPN80DEVICE dev)
{
	int i;
	for (i = 0; i < 10; ++ i)
		dev->keys[i] = 0xff;
}

static void calc_clk_sync(LPN80DEVICE dev, DWORD hz, DWORD interval) {
	dev->param.hz = hz;
	dev->param.sync_interval = interval;

	dev->param.sync_ticks = (UINT) floor(((double) dev->param.hz) * dev->param.sync_interval / 1000.0 + 0.5);
}

export void _n80_device_initialize(LPN80DEVICE dev, SYNCFNC enter, SYNCFNC leave, CMODFNC chgmod)
{
	// clear parameters
	memset(dev, 0, sizeof(N80DEVICE));

	// set methods
	dev->clear_io			= clear_io;
	dev->clear_ram			= clear_ram;
	dev->clear_keys			= clear_keys;
	dev->calc_clk_sync		= calc_clk_sync;
	dev->attr.chg_converter = _n80_chg_converter;	// ATTRIB.c
	dev->attr.invalid		= _n80_attr_invalid;
	dev->attr.set			= _n80_attr_set;
	dev->attr.cache			= _n80_attr_cache;
	dev->attr.eval			= _n80_eval_attrib;
	dev->calender.init		= _n80_init_calender;	// PC1990.c
	dev->calender.mode		= _n80_time_mode;
	dev->crtc.mode			= _n80_width_mode;		// CRTC.c
	dev->crtc.cmd			= _n80_crtc_cmd;
	dev->crtc.data			= _n80_crtc_data;
	dev->cursor.enable		= _n80_cursor_enable;	// CURSOR.c
	dev->cursor.disable		= _n80_cursor_disable;
	dev->cursor.move		= _n80_move_cursor;
	dev->vrtc.edge			= _n80_renew_vrtc;		// VRTC.c
	dev->vrtc.poll			= _n80_poll_vrtc;
	dev->cmt.read			= _n80_cmt_read;		// CMT.c
	dev->cmt.cmd			= _n80_cmt_command;
	dev->hal.mode			= _n80_pcg_mode;		// PCG.c
	dev->vram_sync_enter	= enter;				// critical section API
	dev->vram_sync_leave	= leave;
	dev->chg_crt_mode		= chgmod;

	// call initial methods
	dev->clear_io(&n80device);
	dev->clear_ram(&n80device);
	dev->clear_keys(&n80device);
	dev->calender.init();
	dev->attr.chg_converter(1);						// attribute convert mode init (mono)
	dev->calc_clk_sync(&n80device, 3993600, 20);	// CLK = 4MHz, sync interval = 20ms

	// set default params
	dev->param.flips = 60;				// flip rate = 60fps
	dev->crtc._attr_reverse     = 0;
	dev->crtc._n80_color_mode   = 1;
	dev->crtc._n80_line_mode    = 1;
	dev->cursor._n80_cur_enable = 1;
	dev->cursor._n80_cur_blink  = 1;
}

#ifdef __cplusplus
}	// extern "C"
#endif
