/**--------------------------------------------------------------------------
 **
 **		Z80 Emulator Ver 3.00
 **		Copyright (C)1993-2010 H.Kanemoto
 **
 **		This is a free software. You may copy or distribute
 **		it without any permissions. But there are NO WARRANTY
 **		of use. Entire risk of use is with you.
 **
 **------------------------------------------------------------------------*/
#include "z80.h"
#include "z80mac.h"
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <memory.h>
#ifdef _MSC_VER
#include <crtdbg.h>
#include <mmsystem.h>
#include <windows.h>
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
/**--------------------------------------------------------------------------
 **		Z80 Decoder
 **------------------------------------------------------------------------*/
extern void _z80_decoder(Z80CPU *cpu);
/**--------------------------------------------------------------------------
 **		Z80
 **------------------------------------------------------------------------*/
static int __cdecl __z80_step(LPZ80CPU cpu)
{
	cpu->r.x.T = 0;					// ステートクリア
	cpu->r.x.W = 0;					// メモリウェイトクリア
	cpu->r.x.head = cpu->r.x.PC;	// 現在のインストラクションポインタ退避
	++cpu->clk;						// R register用クロック

	_z80_decoder(cpu);				// １命令実行

#ifdef _ZAURUS_
	cpu->Ts += cpu->r.x.T;
#endif

	if (cpu->signal)				// HALT/未実装/無効命令検知
		return 1;
	else
		return 0;
}

static void __cdecl __z80_reset(LPZ80CPU cpu)
{
	clock_t	t;
	int		i;

	t = clock();					// R register乱数配列
	srand((int) t);
	for (i = 0; i < 256; ++i)
		cpu->rreg[i] = (BYTE) (rand() % 256);

	memset(&cpu->r, 0, sizeof(Z80REGS));
	cpu->signal = 0;
	cpu->clk = 0;
}

export void __z80_set_external_method(LPZ80CPU cpu, LPZ80FNCMBR mbr, LPZ80FNCMBW mbw, LPZ80FNCIOR ior, LPZ80FNCIOW iow)
{
	cpu->step = __z80_step;
	cpu->reset = __z80_reset;

	cpu->mbread = mbr;		// external Memory access function
	cpu->mbwrite = mbw;
	cpu->ioread = ior;		// external I/O access function
	cpu->iowrite = iow;

	__z80_reset(cpu);
}

#ifdef __cplusplus
}	// extern "C"
#endif
